package fr.amap.lidar.amapvox.voxelisation;

import fr.amap.commons.math.util.MatrixUtility;
import fr.amap.commons.raster.asc.Raster;
import fr.amap.commons.util.Cancellable;
import fr.amap.commons.util.IteratorWithException;
import fr.amap.commons.util.TimeCounter;
import fr.amap.commons.util.filter.Filter;
import fr.amap.lidar.amapvox.commons.GTheta;
import fr.amap.lidar.amapvox.commons.LADParams;
import fr.amap.lidar.amapvox.commons.LeafAngleDistribution;
import fr.amap.lidar.amapvox.shot.Shot;
import fr.amap.lidar.amapvox.jeeb.archimed.raytracing.geometry.LineElement;
import fr.amap.lidar.amapvox.jeeb.archimed.raytracing.geometry.LineSegment;
import fr.amap.lidar.amapvox.jeeb.archimed.raytracing.voxel.VoxelManager.VoxelCrossingContext;
import fr.amap.lidar.amapvox.commons.Voxel;
import fr.amap.lidar.amapvox.commons.VoxelSpaceInfos;
import fr.amap.lidar.amapvox.jeeb.archimed.raytracing.util.BoundingBox3d;
import fr.amap.lidar.amapvox.jeeb.archimed.raytracing.voxel.Scene;
import fr.amap.lidar.amapvox.jeeb.archimed.raytracing.voxel.VoxelManager;
import fr.amap.lidar.amapvox.jeeb.archimed.raytracing.voxel.VoxelManagerSettings;
import fr.amap.lidar.amapvox.util.Util;
import fr.amap.lidar.amapvox.voxelisation.configuration.VoxelizationCfg;
import fr.amap.lidar.amapvox.voxelisation.configuration.params.EchoesWeightByFileParams;
import fr.amap.lidar.amapvox.voxelisation.configuration.params.VoxelParameters;
import fr.amap.lidar.amapvox.voxelisation.output.GroundEnergyOutput;
import fr.amap.lidar.amapvox.voxelisation.output.LAIRasterOutput;
import fr.amap.lidar.amapvox.voxelisation.output.NetCDFVoxelOutput;
import fr.amap.lidar.amapvox.voxelisation.output.TxtVoxelOutput;
import fr.amap.lidar.amapvox.voxelisation.postproc.NaNsCorrection;
import java.io.File;
import java.io.IOException;
import java.math.BigInteger;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.List;
import javax.vecmath.Matrix4d;
import javax.vecmath.Point3i;

import javax.vecmath.Point3d;
import javax.vecmath.Vector3d;
import org.apache.log4j.Logger;

public class Voxelization extends fr.amap.commons.util.Process implements Cancellable {

    private final static Logger LOGGER = Logger.getLogger(Voxelization.class);
    private final static double EPSILON = 1e-5;
    private final static double SHOT_MAX_DISTANCE = 10000;

    private boolean cancelled;

    private int nShotsProcessed;
    private int nShotsDiscarded;
    private int nShots;
    private int nShotsOut;

    private Voxel voxels[][][];
    private VoxelManager voxelManager;
    private VoxelManager subVoxelManager;
    private int subVoxelSplit;
    private boolean subSamplingEnabled;

    private float MAX_PAD;

    private double[][] weightTable;
    private EchoesWeightByFileParams echoWeightCorrection;

    private boolean constantBeamSection;

    private boolean groundEnergyEnabled;
    private GroundEnergy[][] groundEnergy;

    private final VoxelizationCfg cfg;
    private final VoxelParameters parameters;
    private Raster dtm;

    private LaserSpecification laserSpec;

    private final List<Filter<Shot>> shotFilters;
    private final List<Filter<Shot.Echo>> echoFilters;
    private EchoesContext currentEchoesContext;

    //directional transmittance (GTheta)
    private GTheta direcTransmittance;

    private boolean padEnabled;
    private boolean approxTransmEnabled;
    private double err_tr_min;
    //private int nTrRecordMax;
    private boolean numEstimTransmEnabled;
    private boolean angleEnabled;
    private boolean lmeanEnabled;
    private boolean ltotEnabled;
    private boolean attenuationEnabled;

    private final List<File> outputFiles;

    private boolean wdistLaserEnabled;
    private boolean sdLengthEnabled;

    private double lambda1;

    public Voxelization(VoxelizationCfg cfg) throws Exception {

        // voxelisation parameters
        this.cfg = cfg;
        this.parameters = cfg.getVoxelParameters();

        // user defined shot filters
        shotFilters = cfg.getShotFilters();

        // user defined echo filters
        this.echoFilters = cfg.getEchoFilters();

        outputFiles = new ArrayList();
    }

    public File[] voxelization(IteratorWithException<Shot> it) throws Exception {

        fireProgress("Voxelisation", 0, 100);

        long startTime = System.currentTimeMillis();

        init();

        while (it.hasNext()) {
            // check for cancellation
            if (isCancelled()) {
                LOGGER.info("Voxelization cancelled (" + cfg.getInputFile().getName() + ")");
                return null;
            }
            processOneShot(it.next());
        }

        LOGGER.info("[voxelisation] Shots summary: processed " + nShotsProcessed + ", discarded " + nShotsDiscarded + ", out " + nShotsOut);

        postProcess();

        if (isCancelled()) {
            return null;
        }

        writeOutput();

        fireFinished(TimeCounter.getElapsedTimeInSeconds(startTime));

        return outputFiles.toArray(new File[outputFiles.size()]);
    }

    public List<Filter<Shot>> getShotFilters() {
        return shotFilters;
    }

    public void processOneShot(Shot shot) throws Exception {

        if (voxelManager == null) {
            LOGGER.error("VoxelManager not initialized, what happened??");
            return;
        }

        if ((nShots % 1000000) == 0) {
            printVoxelisationProgress();
        }
        nShots++;

        if (retainShot(shot)) {
            // vegetation free shot propagation (as if no vegetation in the scene)
            freePropagation(shot);

            // shot propagation (with light interception in the scene)
            if (propagation(shot, currentEchoesContext)) {
                // increment number of shots processed
                nShotsProcessed++;
            } else {
                nShotsOut++;
            }
        } else {
            nShotsDiscarded++;
        }
    }

    /**
     * Propagate a {@code shot} in the voxel space ignoring the vegetation in
     * order to estimate a potential beam volume.
     *
     * @param shot
     */
    private void freePropagation(Shot shot) {

        LineElement shotLine = shotLine(shot);

        // first voxel crossed by the shot
        VoxelCrossingContext context = voxelManager.getFirstVoxelV2(shotLine);

        if (null != context) {
            do {
                // current voxel
                Point3i vcoord = context.indices;

                // stop propagation if current voxel is below the ground
                if (belowGround(voxels[vcoord.x][vcoord.y][vcoord.z])) {
                    break;
                }

                // compute beam surface at voxel centre
                double beamSurface = constantBeamSection
                        ? 1.d
                        : beamSection(shot.origin, vcoord, laserSpec);

                // increment potential beam volume
                voxels[vcoord.x][vcoord.y][vcoord.z].bsPotential += beamSurface;

                // get next voxel
                context = voxelManager.CrossVoxel(shotLine, vcoord);

                // loop until exiting voxel space
            } while (context.indices != null);
        }
    }

    /**
     * Propagate a Shot inside a given voxel and assess the sampling at sub
     * voxel scale.
     *
     * @param segment, the shot portion to propagate
     * @param vcoord
     * @return a BigInteger with sampling intensity at sub voxel scale.
     */
    private BigInteger subPropagation(Shot shot, Point3i vcoord) {

        LineSegment shotLine = shotLine(shot);
        Vector3d normVec = new Vector3d(voxelManager.getInfCorner(vcoord));
        normVec.negate();
        shotLine.translate(normVec);

        BigInteger subsampling = BigInteger.ZERO;

        // first sub voxel intersected by the shot
        VoxelCrossingContext voxelCrossing = subVoxelManager.getFirstVoxelV2(shotLine);
        if (null != voxelCrossing) {
            do {
                // current sub voxel index
                Point3i subvcoord = voxelCrossing.indices;
                // increment sub voxel sampling
                subsampling = subsampling.setBit(subijkToIndex(subvcoord));
                // next subvoxel
                voxelCrossing = subVoxelManager.CrossVoxel(shotLine, subvcoord);
                // loop until exiting voxel
            } while (null != voxelCrossing.indices);
        }

        return subsampling;
    }

    private int subijkToIndex(Point3i vcoord) {
        return vcoord.z + subVoxelSplit * (vcoord.y + subVoxelSplit * vcoord.x);
    }

    private LineSegment shotLine(Shot shot) {
        return new LineSegment(shot.origin, shot.direction, SHOT_MAX_DISTANCE);
    }

    private boolean propagation(Shot shot, EchoesContext echoesContext) throws Exception {

        // create shot line
        LineSegment shotLine = shotLine(shot);

        // first voxel intersected by the shot, refered as voxel0 in further comments
        VoxelCrossingContext voxelCrossing = voxelManager.getFirstVoxelV2(shotLine);

        if (null != voxelCrossing) {
            // initialise entering beam fraction (100%)
            double beamFractionIn = 1.d;
            // look for first echo inside or beyond voxel0 (if any)
            int rank = 0;
            // by default first echo is null
            Shot.Echo echo = null;
            if (echoesContext.nEchoes > 0) {
                // look for echoes located before voxel0
                while ((rank < echoesContext.nEchoes) && (shot.getRange(rank) < voxelCrossing.length)) {
                    // decrement beam fraction entering voxel0
                    beamFractionIn -= echoesContext.bfIntercepted[rank];
                    rank++;
                }

                if (beamFractionIn < EPSILON) {
                    // no light reaches voxel0 => no propagation inside voxel space
                    return false;
                } else if (rank < echoesContext.nEchoes) {
                    // there is at list one echo located inside or beyond voxel0
                    echo = shot.echoes[rank];
                }
                // all the echoes are located before voxel0 but beamFractionIn
                // reaching voxel0 not zero => propagation with echo = null
            }

            // loop over the voxels crossed by the shot
            Point3i vcoord;
            boolean interruptPropagation;
            do {
                // current voxel coordinates
                vcoord = voxelCrossing.indices;
                int i = vcoord.x;
                int j = vcoord.y;
                int k = vcoord.z;

                // stop propagation if current voxel is below the ground
                if (!groundEnergyEnabled && belowGround(voxels[i][j][k])) {
                    break;
                }

                // ray length within voxel
                // distance from shot origin to shot interception point with current voxel
                double dIn = voxelCrossing.length;
                // get next voxel
                voxelCrossing = voxelManager.CrossVoxel(shotLine, vcoord);
                // distance from shot origin to shot interception point with next voxel
                double dOut = voxelCrossing.length;
                double rayLengthVoxel = dOut - dIn;

                // reset intercepted beam fraction in current voxel
                double bfIntercepted = 0.d;

                // initializes list of ray records for current voxel
                List<RayRecord> rayRecords = new ArrayList();

                // handles echoes that fall inside current voxel
                if (null != echo && isEchoInsideVoxel(echo, vcoord)) {

                    // loop over echoes in same voxel
                    while ((null != echo) && isEchoInsideVoxel(echo, vcoord) && (beamFractionIn - bfIntercepted > EPSILON)) {
                        // add ray associated to current echo (intercepted)
                        rayRecords.add(new RayRecord(shot.getRange(rank) - dIn, echoesContext.bfIntercepted[rank], true, echoesContext.retained[rank]));
                        // intercepted beam fraction regardless of echo filtering
                        bfIntercepted += echoesContext.bfIntercepted[rank];
                        // next echo 
                        rank++;
                        echo = rank < echoesContext.nEchoes ? shot.echoes[rank] : null;
                    }
                }
                // add ray that goes through voxel (not intercepted)
                if (beamFractionIn - bfIntercepted > EPSILON) {
                    rayRecords.add(new RayRecord(rayLengthVoxel, beamFractionIn - bfIntercepted));
                }

                // update voxel state variables
                updateVoxel(vcoord, shot, rayLengthVoxel, rayRecords);

                // increment ground energy
                if (groundEnergyEnabled && closeToGround(voxels[i][j][k])) {
                    groundEnergy[i][j].groundEnergyPotential++;
                    groundEnergy[i][j].groundEnergyActual += beamFractionIn;
                }

                // decrement beamFractionIn for next voxel
                beamFractionIn -= bfIntercepted;
                // check whether the propagation should be interrupted
                //   residual beam fraction is null
                //   reached edge of the voxel space
                //   mono-echo laser and first echo encountered already
                interruptPropagation = (beamFractionIn < EPSILON)
                        || (null == voxelCrossing.indices)
                        || (laserSpec.isMonoEcho() && (null != echo) && (echo.rank > 0));

            } while (!interruptPropagation);
            // shot went through the voxel space
            return true;
        }
        // shot did not go through the voxel space
        return false;
    }

    /**
     * Update voxel state variables based on ray records.
     *
     * @param index, the voxel index (i, j, k)
     * @param shot, current shot
     * @param rayLenghVoxel, potential ray length (meter) in current voxel, from
     * entering point to exiting point.
     * @param rayRecords, a list of ray records of current shot in current voxel
     */
    private void updateVoxel(Point3i index, Shot shot, double rayLengthVoxel, List<RayRecord> rayRecords) {

        int i = index.x;
        int j = index.y;
        int k = index.z;

        // compute beam surface at voxel centre
        double beamSurface = constantBeamSection
                ? 1.d
                : beamSection(shot.origin, index, laserSpec);

        // increment number of shots crossing current voxel
        voxels[i][j][k].nbSampling++;
        // initializes entering beam surface and cumulated ray length
        double bsEntering = 0.d;
        double bsIntercepted = 0.d;
        double rayLength = 0.d;
        int nray = 0;
        // loop over ray records
        for (RayRecord raySegment : rayRecords) {
            double beamSection = raySegment.beamFraction * beamSurface;
            // increment entering beam fraction in current voxel
            bsEntering += beamSection;
            if (raySegment.intercepted && raySegment.retained) {
                bsIntercepted += beamSection;
                // increment number of echoes
                voxels[i][j][k].nbEchos++;
                // increment intercepted beam section
                voxels[i][j][k].bsIntercepted += beamSection;
                // increment intercepted beam volume (F. Pimont's method)
                if (attenuationEnabled) {
                    voxels[i][j][k].attenuationVar.interceptedBV += beamSection * opticalLength(raySegment.length);
                }
            } else {
                // here we consider the beam fraction that either goes
                // through the whole voxel or that is intercepted but
                // discarded (rulled out as wood for instance).
                // increment potential beam volume (F. Pimont's method)
                if (attenuationEnabled) {
                    voxels[i][j][k].attenuationVar.potentialBV += beamSection * opticalLength(raySegment.length);
                }
                rayLength += raySegment.length;
                nray += 1;
            }
        } //  end of loop over ray records
        voxels[i][j][k].bsEntering += bsEntering;
        // increment total optical length in current voxel
        // phv 20190713 - trick for handling rays that are intercepted 
        // but discarded (as wood echo for instance). In such case
        // we consider that the intercepted beam fraction goes through
        // with optical length from voxel entering point to echo.
        // So if it is a last echo lgTotal will be incremented by 
        // distance from voxel entering point to echo. If it is not the
        // last echo lgTotal will be incremented with the average of
        // the distance from voxel entering point to echo (the discarded
        // beam fraction) and the distance from voxel entering point to
        // voxel exiting point (the crossing beam fraction).
        rayLength = (nray > 0) ? rayLength / nray : rayLengthVoxel;
        if (ltotEnabled) {
            voxels[i][j][k].lgTotal += rayLength;
        }
        // increment mean angle in current voxel
        if (angleEnabled) {
            voxels[i][j][k].angleMean += shot.getAngle();
        }
        // update variables for numerical estimation of the transmittance
        if (numEstimTransmEnabled) {
            int nTrEstim = voxels[i][j][k].trNumBsOut.length;
            for (int itr = 0; itr < nTrEstim; itr += 1) {
                double tr =  err_tr_min * (itr + 1);
                voxels[i][j][k].trNumBsOut[itr] += bsEntering * Math.pow(tr, rayLength);
            }
        }
        // update mean ray length and associated standard deviation
        if (sdLengthEnabled) {
            // Welford's method for computing variance
            double oldMeanLength = voxels[i][j][k].lMeanTotal;
            voxels[i][j][k].lMeanTotal += (rayLength - oldMeanLength) / voxels[i][j][k].nbSampling;
            voxels[i][j][k].sdLength += (rayLength - oldMeanLength) * (rayLength - voxels[i][j][k].lMeanTotal);
        }
        // increment distance to laser
        if (wdistLaserEnabled) {
            voxels[i][j][k].distLaser += voxelManager.getCenter(index).distance(shot.origin);
        }
        // sub voxel propagation
        if (subSamplingEnabled) {
            voxels[i][j][k].subSampling = voxels[i][j][k].subSampling.or(subPropagation(shot, index));
        }
    }

    private boolean isEchoInsideVoxel(Shot.Echo echo, Point3i indexVoxel) {

        Point3i indexEcho = findVoxel(echo);

        return (null != indexEcho) && (indexEcho.equals(indexVoxel));
    }

    private double beamAttenuation(int rankEcho, int nEcho, double weightCorr) {

        // compute beam fraction of current echo
        double beamFractionCurrentEcho = 1;
        // beam fraction pondered by weight table
        if (null != weightTable && (nEcho > 0)) {
            beamFractionCurrentEcho *= weightTable[nEcho - 1][rankEcho];
        }
        // beam fraction pondered by weight from CSV file
        beamFractionCurrentEcho *= weightCorr;
        return beamFractionCurrentEcho;
    }

    /**
     * Compute the average beam section of the shot in given voxel by estimating
     * it at voxel centre.
     *
     * @param origin, the shot origin
     * @param index, the voxel index (i, j, k)
     * @param spec, laser specification (beam divergence and diameter at exit)
     * @return the beam section calculated at voxel centre.
     */
    private double beamSection(Point3d origin, Point3i index, LaserSpecification spec) {

        // distance from shot origin to current voxel center
        Point3d voxelPosition = getPosition(index);
        double distance = voxelPosition.distance(origin);

        // beam surface in current voxel
        return Math.pow((Math.tan(0.5d * spec.getBeamDivergence()) * distance) + 0.5d * spec.getBeamDiameterAtExit(), 2) * Math.PI;
    }

    /**
     * check whether current voxel centre vertical coordinate is below the
     * ground
     */
    private boolean belowGround(Voxel voxel) {
        return (voxel.ground_distance < voxelManager.getVoxelSpace().getVoxelSize().z / 2.0f);
    }

    private boolean closeToGround(Voxel voxel) {
        return (voxel.ground_distance < parameters.getDtmFilteringParams().getMinDTMDistance());
    }

    private EchoesContext extractEchoesContext(Shot shot) throws Exception {

        EchoesContext context = new EchoesContext(shot.getEchoesNumber());

        if (context.nEchoes > 0) {
            for (int k = 0; k < context.nEchoes; k++) {
                context.retained[k] = retainEcho(shot.echoes[k]);
            }

            // specific weight attenuation (EchoesWeightByFileParams.java) for this shot
            double weightCorr = (null != echoWeightCorrection)
                    ? echoWeightCorrection.getWeightCorrection(shot.index)
                    : 1.d;
            for (int k = 0; k < context.nEchoes; k++) {
                context.bfIntercepted[k] = beamAttenuation(k, context.nEchoes, weightCorr);
            }
        }

        return context;
    }

    public void printVoxelisationProgress() {
        LOGGER.info("[voxelisation] shot " + nShots + " (processed " + nShotsProcessed + ", discarded " + nShotsDiscarded + ", out " + nShotsOut + ")");
    }

    public float getGroundDistance(float x, float y, float z) {

        return (dtm != null && parameters.getDtmFilteringParams().useDTMCorrection())
                ? z - (float) (dtm.getSimpleHeight(x, y))
                : z;
    }

    public void init() throws Exception {

        nShotsProcessed = 0;
        nShotsDiscarded = 0;
        nShots = 0;

        // initialise shot filters
        if (null != shotFilters) {
            for (Filter filter : shotFilters) {
                filter.init();
                LOGGER.info("Initialized shot filter " + filter.getClass().getSimpleName());
            }
        }

        // initialise echo filters
        if (null != echoFilters) {
            for (Filter filter : echoFilters) {
                filter.init();
                LOGGER.info("Initialized echo filter " + filter.getClass().getSimpleName());
            }
        }

        // echo weighting table with security check to ensure values <= 1
        if (null != parameters.getEchoesWeightByRankParams()) {
            weightTable = parameters.getEchoesWeightByRankParams().getWeightingData();
            for (int i = 0; i < weightTable.length; i++) {
                for (int j = 0; j < weightTable[i].length; j++) {
                    if (!Double.isNaN(weightTable[i][j]) && weightTable[i][j] > 1) {
                        throw new IOException("Inconsistent echo weighting table with values greater than one. Please fix it in the configuration file.");
                    }
                    if (sum(weightTable[i]) > 1.d) {
                        throw new IOException("Inconsistent echo weighting table. Cumulated values at line " + (i + 1) + " greater than one. " + Arrays.toString(weightTable[i]));
                    }
                }
            }
        } else {
            // no energy ponderation by rank
            weightTable = new double[7][7];
            for (int i = 0; i < weightTable.length; i++) {
                for (int j = 0; j < weightTable[i].length; j++) {
                    weightTable[i][j] = j < (i + 1) ? 1.d : Double.NaN;
                }
            }
            // warn user that current algorithm, without ponderation table, will
            // behave as mono-echo voxelisation even though the scan is multi-echo
            LOGGER.warn("No echo weighting table provided. AMAPVox will voxelize the scans as mono-echo (even though they are multi-echoes).");
        }

        if (null != parameters.getEchoesWeightByFileParams()) {
            echoWeightCorrection = parameters.getEchoesWeightByFileParams();
            echoWeightCorrection.init();
        }

        MAX_PAD = parameters.infos.getMaxPAD();

        laserSpec = parameters.getLaserSpecification();

        LADParams ladParameters = parameters.getLadParams();
        if (ladParameters == null) {
            ladParameters = new LADParams();
        }

        LeafAngleDistribution distribution = new LeafAngleDistribution(ladParameters.getLadType(),
                ladParameters.getLadBetaFunctionAlphaParameter(),
                ladParameters.getLadBetaFunctionBetaParameter());

        direcTransmittance = new GTheta(distribution);

        LOGGER.info("Building transmittance functions table");
        direcTransmittance.buildTable(GTheta.DEFAULT_STEP_NUMBER);
        LOGGER.info("Transmittance functions table is built");

        groundEnergyEnabled = parameters.getGroundEnergyParams() != null
                && parameters.getGroundEnergyParams().isCalculateGroundEnergy()
                && parameters.infos.getType() != VoxelSpaceInfos.Type.TLS;

        // enabled output variables
        padEnabled = parameters.isOutputVariableEnabled(Voxel.OutputVariable.PLANT_AREA_DENSITY);
        numEstimTransmEnabled = padEnabled || parameters.isOutputVariableEnabled(Voxel.OutputVariable.ESTIMATED_TRANSMITTANCE);
        approxTransmEnabled = parameters.isOutputVariableEnabled(Voxel.OutputVariable.APPROXIMATED_TRANSMITTANCE);
        angleEnabled = padEnabled || parameters.isOutputVariableEnabled(Voxel.OutputVariable.MEAN_ANGLE);
        lmeanEnabled = approxTransmEnabled || parameters.isOutputVariableEnabled(Voxel.OutputVariable.MEAN_TOTAL_LENGTH);
        ltotEnabled = lmeanEnabled || parameters.isOutputVariableEnabled(Voxel.OutputVariable.TOTAL_LENGTH);
        attenuationEnabled = parameters.isOutputVariableEnabled(Voxel.OutputVariable.ATTENUATION_COEFF);

        err_tr_min = parameters.getTrNumEstimError();
        //nTrRecordMax = parameters.getNTrRecordMax();

        wdistLaserEnabled = parameters.isOutputVariableEnabled(Voxel.OutputVariable.DIST_LASER);
        sdLengthEnabled = parameters.isOutputVariableEnabled(Voxel.OutputVariable.SD_LENGTH);

        // digital terrain model 
        if (cfg.getVoxelParameters().getDtmFilteringParams().useDTMCorrection()) {
            LOGGER.info("Loading dtm...");
            dtm = Util.loadDTM(cfg.getVoxelParameters().getDtmFilteringParams().getDtmFile());
            if (dtm != null && cfg.getVoxelParameters().getDtmFilteringParams().isUseVOPMatrix()) {
                Matrix4d vop = (null == cfg.getVopMatrix())
                ? MatrixUtility.identity4d()
                : new Matrix4d(cfg.getVopMatrix());
                dtm.setTransformationMatrix(vop);
            }
        }

        // sub voxel space exploration
        subSamplingEnabled = parameters.isOutputVariableEnabled(Voxel.OutputVariable.EXPLORATION_RATE);
        if (subSamplingEnabled) {
            subVoxelSplit = parameters.getSubVoxelSplit();
        }

        // create voxel space
        createVoxelSpace();

        // unitary extinction coefficient lambda1
        Point3d voxSize = voxelManager.getVoxelSpace().getVoxelSize();
        lambda1 = 0.25d * parameters.getMeanLeafArea() / (voxSize.x * voxSize.y * voxSize.z);

        outputFiles.clear();
    }

    /**
     * Get position of the center of a voxel
     *
     * @param vcoord, voxel coordinate
     * @return
     */
    public Point3d getPosition(Point3i vcoord) {

        Point3d minCorner = parameters.infos.getMinCorner();
        Point3d voxSize = voxelManager.getVoxelSpace().getVoxelSize();

        double posX = minCorner.x + (voxSize.x / 2.0d) + (vcoord.x * voxSize.x);
        double posY = minCorner.y + (voxSize.y / 2.0d) + (vcoord.y * voxSize.y);
        double posZ = minCorner.z + (voxSize.z / 2.0d) + (vcoord.z * voxSize.z);

        return new Point3d(posX, posY, posZ);
    }

    /**
     * Check whether given 3d point falls right on a voxel facet.
     *
     * @param point3d
     * @return true if the point falls on a voxel facet, false if strictly
     * inside voxel.
     */
    private boolean[] isOnVoxelFacet(Point3d point3d) {

        Point3d size = voxelManager.getVoxelSpace().getVoxelSize();
        boolean[] onFacet = new boolean[3];
        if (voxelManager.isPointInsideBoundingBox(point3d, false)) {
            onFacet[0] = Math.abs(point3d.x % size.x) < EPSILON;
            onFacet[1] = Math.abs(point3d.y % size.y) < EPSILON;
            onFacet[2] = Math.abs(point3d.z % size.z) < EPSILON;
        }
        return onFacet;
    }

    /**
     * Find the voxel that contains the given echo, taking into account specific
     * boundary condition. Boundary condition: for an echo that falls on a facet
     * shared by two voxels AMAPVox assumes it belongs to the voxel in which the
     * optical ray is not zero, taking into account the shot trajectory. E.g.:
     * shot going along y = 1, echo y = 3 falls in between voxel 2 and 3.
     * Mathematically speaking it belongs to voxel 3 but in such case the
     * optical length in voxel 3 would be zero so AMAPVox assumes it belongs to
     * voxel 2 with an optical length equal to 1.
     *
     * @param echo
     * @return the index of the voxel containing the echo.
     */
    private Point3i findVoxel(Shot.Echo echo) {

        Point3i vcoord = voxelManager.getVoxelIndicesFromPoint(echo.location);
        if (null != vcoord) {
            boolean[] onFacet = isOnVoxelFacet(echo.location);
            if (onFacet[0] && Math.signum(echo.shot.direction.x) > 0) {
                //LOGGER.info("Special boundary condition on x " + echo.location);
                vcoord.x = (int) Math.max(vcoord.x - 1, 0);
            }
            if (onFacet[1] && Math.signum(echo.shot.direction.y) > 0) {
                //LOGGER.info("Special boundary condition on y " + echo.location);
                vcoord.y = (int) Math.max(vcoord.y - 1, 0);
            }
            if (onFacet[2] && Math.signum(echo.shot.direction.z) > 0) {
                //LOGGER.info("Special boundary condition on z " + echo.location);
                vcoord.z = (int) Math.max(vcoord.z - 1, 0);
            }
        }
        return vcoord;
    }

    boolean retainEcho(Shot.Echo echo) throws Exception {

        if (echo.rank >= 0 && echoFilters != null) {
            for (Filter filter : echoFilters) {
                if (!filter.accept(echo)) {
                    return false;
                }
            }
        }

        // DTM filtering
        if (parameters.getDtmFilteringParams().useDTMCorrection()) {
            float echoDistance = getGroundDistance((float) echo.location.x, (float) echo.location.y, (float) echo.location.z);
            if (!Float.isNaN(echoDistance)) {
                return (echoDistance >= parameters.getDtmFilteringParams().getMinDTMDistance());
            }
        }

        // echo retained by every filter
        return true;
    }

    boolean retainShot(Shot shot) throws Exception {

        if (null != shotFilters) {
            for (Filter<Shot> filter : shotFilters) {
                // as soon as a filter discard the shot returns false 
                if (!filter.accept(shot)) {
                    return false;
                }
            }
        }

        // echo filters and weighting
        try {
            currentEchoesContext = extractEchoesContext(shot);
        } catch (Exception ex) {
            // warn the user and discard the shot
            LOGGER.warn(ex);
            return false;
        }

        // all filters retain the shot
        return true;
    }

    boolean isInsideSameVoxel(Shot.Echo echo1, Shot.Echo echo2) {

        if (echo1 == null || echo2 == null) {
            return false;
        }

        Point3i vcoord1 = findVoxel(echo1);
        Point3i vcoord2 = findVoxel(echo2);

        return vcoord1 != null && vcoord2 != null && vcoord1.equals(vcoord2);
    }

    public static double approximateTransmittance(double bfEntering, double bfIntercepted, double lMeanTotal) {

        // unsampled voxel
        if ((bfEntering == 0)) {
            return Double.NaN;
        }

        // no light interception, transmittance = 1
        if (bfIntercepted == 0) {
            return 1.d;
        }

        double bfOut = bfEntering - bfIntercepted;

        // no light transmitted, transmittance = 0
        if (Math.abs(bfOut) < EPSILON) {
            return 0.d;
        }

        return Math.pow(bfOut / bfEntering, 1 / lMeanTotal);
    }

    public static double computePADFromNormTransmittance(double transmittance, double angleMean, double maxPAD, GTheta direcTransmittance) {

        double pad;

        if (Double.isNaN(transmittance)) {
            pad = Double.NaN;
        } else if (transmittance == 0) {
            pad = maxPAD;
        } else {
            double coefficientGTheta = direcTransmittance.getGThetaFromAngle(angleMean, true);
            pad = Math.log(transmittance) / (-coefficientGTheta);
            if (Double.isNaN(pad)) {
                pad = Double.NaN;
            } else if (pad > maxPAD || Double.isInfinite(pad)) {
                pad = maxPAD;
            }
        }

        return pad + 0.0d; //set +0.0f to avoid -0.0f
    }

    private double computePADFromNormTransmittance(double transmittance, double angleMean) {

        return computePADFromNormTransmittance(transmittance, angleMean, MAX_PAD, direcTransmittance);
    }

    public void postProcess() {

        LOGGER.info("Post-processing (computing transmittance, plant area density, etc.). Please wait...");

        double subSamplingMax = Math.pow(subVoxelSplit, 3);
        for (int i = 0; i < parameters.infos.getSplit().x; i++) {
            for (int j = 0; j < parameters.infos.getSplit().y; j++) {
                for (int k = 0; k < parameters.infos.getSplit().z; k++) {
                    if (isCancelled()) {
                        LOGGER.info("Post-processing cancelled");
                        return;
                    }
                    // empty voxel, no post-processing
                    if (voxels[i][j][k].nbSampling == 0) {
                        voxels[i][j][k].angleMean = Double.NaN;
                        voxels[i][j][k].approxTransm = Double.NaN;
                        voxels[i][j][k].PadBVTotal = Double.NaN;
                        if (numEstimTransmEnabled) {
                            voxels[i][j][k].transmittance = Double.NaN;
                        }
                        if (wdistLaserEnabled) {
                            voxels[i][j][k].distLaser = Double.NaN;
                        }
                        voxels[i][j][k].attenuation_biasCorr = Double.NaN;
                        voxels[i][j][k].attenuation = Double.NaN;
                        if (subSamplingEnabled) {
                            voxels[i][j][k].explorationRate = Double.NaN;
                        }
                        continue;
                    }

                    Voxel voxel = voxels[i][j][k];

                    // mean angle
                    if (angleEnabled) {
                        voxels[i][j][k].angleMean = voxel.angleMean / voxel.nbSampling;
                    }

                    // mean optical length
                    if (lmeanEnabled) {
                        voxels[i][j][k].lMeanTotal = voxel.lgTotal / voxel.nbSampling;
                    }

                    if (sdLengthEnabled) {
                        // phv 20190617 Welford’s method for computing variance (and sd)
                        voxels[i][j][k].sdLength = voxels[i][j][k].nbSampling > 1
                                ? Math.sqrt(voxels[i][j][k].sdLength / (voxels[i][j][k].nbSampling - 1))
                                : 0.d;
                    }

                    // approximated transmittance
                    if (approxTransmEnabled) {
                        double normalizedTransmittance = approximateTransmittance(voxel.bsEntering, voxel.bsIntercepted, voxel.lMeanTotal);
                        voxels[i][j][k].approxTransm = (float) normalizedTransmittance;
                    }

                    // numerical estimation of the transmittance
                    if (numEstimTransmEnabled) {
                        voxels[i][j][k].transmittance = estimateTransmittance(
                                voxels[i][j][k].bsEntering,
                                voxels[i][j][k].bsIntercepted,
                                voxels[i][j][k].trNumBsOut
                        );
                    }

                    // plant area density
                    if (padEnabled) {
                        voxels[i][j][k].PadBVTotal = (float) computePADFromNormTransmittance(voxels[i][j][k].transmittance, voxel.angleMean);
                    }

                    // averaged distance to laser
                    if (wdistLaserEnabled) {
                        voxels[i][j][k].distLaser = voxel.distLaser / voxel.nbSampling;
                    }

                    // attenuation coefficient (F. Pimont's method)
                    if (attenuationEnabled) {
                        double interceptedPlusPotential = voxels[i][j][k].attenuationVar.interceptedBV + voxels[i][j][k].attenuationVar.potentialBV;
                        if (interceptedPlusPotential > 0.d) {
                            voxels[i][j][k].attenuation_biasCorr = voxels[i][j][k].attenuationVar.interceptedBV / Math.pow(interceptedPlusPotential, 2);
                            voxels[i][j][k].attenuation = voxels[i][j][k].bsIntercepted / interceptedPlusPotential;
                        } else {
                            voxels[i][j][k].attenuation_biasCorr = Double.NaN;
                            voxels[i][j][k].attenuation = Double.NaN;
                        }
                    }

                    // sub voxel sampling
                    if (subSamplingEnabled) {
                        voxels[i][j][k].explorationRate = voxels[i][j][k].subSampling.bitCount() / (double) subSamplingMax;
                    }
                }
            }
        }

        if (cfg.getVoxelParameters().getNaNsCorrectionParams().isActivate()) {
            fireProgress("NA correction", 0, 100);
            new NaNsCorrection().correct(parameters, voxels);
        }
    }

    private double estimateTransmittance(double bsEntering, double bsIntercepted, double[] trNumBsOut) {

        double bsOut = bsEntering - bsIntercepted;

        // unsampled voxel
        if ((bsEntering == 0)) {
            return Double.NaN;
        }

        // no light interception, transmittance = 1
        if (bsIntercepted == 0) {
            return 1.d;
        }

        // no light transmitted, transmittance = 0
        if (Math.abs(bsOut) < EPSILON) {
            return 0.d;
        }

        // transmittance associated to minimal error
        double err = Double.MAX_VALUE;
        double tr = Double.NaN;
        int nTrEstim = trNumBsOut.length;
        for (int i = 0; i < nTrEstim; i++) {
            double errTmp = Math.abs(trNumBsOut[i] - bsOut);
            if (errTmp < err) {
                tr = (i + 1) * err_tr_min;
                err = errTmp;
            }
        }
        return tr;
    }

    public void writeOutput() throws IOException, Exception {

        VoxelizationCfg.VoxelsFormat format = cfg.getVoxelsFormat();

        if (null != format) {
            switch (format) {
                case NONE:
                    return;
                case VOXEL:
                    outputFiles.addAll(Arrays.asList(new TxtVoxelOutput(this).write()));
                    break;
                case RASTER:
                    outputFiles.addAll(Arrays.asList(new LAIRasterOutput(this).write()));
                    break;
                case NETCDF:
                    outputFiles.addAll(Arrays.asList(new NetCDFVoxelOutput(this).write()));
                    break;
            }
        }

        if (parameters.getGroundEnergyParams() != null && parameters.getGroundEnergyParams().isCalculateGroundEnergy()) {
            outputFiles.addAll(Arrays.asList(new GroundEnergyOutput(this).write()));
        }
    }

    public void createVoxelSpace() {

        // main voxel space
        voxels = new Voxel[parameters.infos.getSplit().x][parameters.infos.getSplit().y][parameters.infos.getSplit().z];

        if (groundEnergyEnabled) {
            // allocate
            //LOGGER.info("allocate!!!!!!!!");
            groundEnergy = new GroundEnergy[parameters.infos.getSplit().x][parameters.infos.getSplit().y];
            for (int i = 0; i < parameters.infos.getSplit().x; i++) {
                for (int j = 0; j < parameters.infos.getSplit().y; j++) {
                    groundEnergy[i][j] = new GroundEnergy();
                }
            }
        }

        Scene scene = new Scene();
        scene.setBoundingBox(new BoundingBox3d(parameters.infos.getMinCorner(), parameters.infos.getMaxCorner()));

        voxelManager = new VoxelManager(scene, new VoxelManagerSettings(parameters.infos.getSplit(), VoxelManagerSettings.NON_TORIC_FINITE_BOX_TOPOLOGY));

        LOGGER.info(voxelManager.getInformations());

        // initialize voxel
        for (int i = 0; i < parameters.infos.getSplit().x; i++) {
            for (int j = 0; j < parameters.infos.getSplit().y; j++) {
                for (int k = 0; k < parameters.infos.getSplit().z; k++) {
                    voxels[i][j][k] = initVoxel(i, j, k);
                }
            }
        }

        // sub voxel space
        if (subSamplingEnabled) {
            Scene subScene = new Scene();
            subScene.setBoundingBox(new BoundingBox3d(new Point3d(), parameters.infos.getVoxelSize()));
            subVoxelManager = new VoxelManager(subScene, new VoxelManagerSettings(new Point3i(subVoxelSplit, subVoxelSplit, subVoxelSplit), VoxelManagerSettings.NON_TORIC_FINITE_BOX_TOPOLOGY));
        }

    }

    private Voxel initVoxel(int i, int j, int k) {

        Voxel vox = new Voxel(i, j, k);

        Point3d position = getPosition(new Point3i(i, j, k));

        float dist;

        if (dtm != null && parameters.getDtmFilteringParams().useDTMCorrection()) {

            float dtmHeightXY = dtm.getSimpleHeight((float) position.x, (float) position.y);
            if (Float.isNaN(dtmHeightXY)) {
                dist = (float) (position.z);
            } else {
                dist = (float) (position.z - dtmHeightXY);
            }

        } else {
            dist = (float) (position.z - parameters.infos.getMinCorner().z);
        }

        vox.ground_distance = dist;

        if (subSamplingEnabled) {
            vox.subSampling = BigInteger.ZERO;
        }
        
        if (numEstimTransmEnabled) {
            int nTrEstim = (int) (1.d / err_tr_min) - 1;
            vox.trNumBsOut = new double[nTrEstim];
        }

        return vox;
    }

    public Voxel getVoxel(int i, int j, int k) {
        return voxels[i][j][k];
    }

    public int getNbShotsProcessed() {
        return nShotsProcessed;
    }

    public Voxel[][][] getVoxels() {
        return voxels;
    }

    @Override
    public boolean isCancelled() {
        return cancelled;
    }

    @Override
    public void setCancelled(boolean cancelled) {
        this.cancelled = cancelled;
    }

    public void setConstantBeamSection(boolean enabled) {
        this.constantBeamSection = enabled;
    }

    public GroundEnergy[][] getGroundEnergy() {
        return groundEnergy;
    }

    public VoxelizationCfg getVoxelizationCfg() {
        return cfg;
    }

    public VoxelParameters getVoxelParameters() {
        return parameters;
    }

    private double sum(double[] array) {

        double sum = 0.d;
        for (double d : array) {
            if (!Double.isNaN(d)) {
                sum += d;
            }
        }
        return sum;
    }

    private double opticalLength(double rayLength) {
        return -Math.log(1.d - lambda1 * rayLength) / lambda1;
    }

    /**
     *
     */
    public class GroundEnergy {

        /*ATTENTION : le calcul de l'énergie arrivant au sol est incorrect actuellement,
         en effet, la prolongation d'un tir après son dernier écho jusqu'au sol ne se fait
         pas pour les tirs ayant le dernier écho en dehors de l'espace voxel.
         En correction, il faudrait faire un pré-calcul pour déterminer la taille de la 
        bounding-box pour inclure les tirs correspondants.
        Donc créer deux espaces voxels en l'occurence.*/
        public int groundEnergyPotential;
        public float groundEnergyActual;

        public GroundEnergy() {
            groundEnergyPotential = 0;
            groundEnergyActual = 0;
        }
    }

    private class EchoesContext {

        private final int nEchoes;
        private final boolean retained[];
        private final double[] bfIntercepted;

        EchoesContext(int nEchoes) {
            this.nEchoes = nEchoes;
            this.retained = new boolean[nEchoes];
            this.bfIntercepted = new double[nEchoes];
        }
    }

    private class RayRecord {

        private final double length;
        private final double beamFraction;
        private final boolean intercepted;
        private final boolean retained;

        RayRecord(double length, double beamFraction, boolean intercepted, boolean retained) {
            this.length = length;
            this.beamFraction = beamFraction;
            this.intercepted = intercepted;
            this.retained = retained;
        }

        RayRecord(double length, double beamFraction) {
            this(length, beamFraction, false, true);
        }
    }
}
